// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright 2015 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/
/*  ̃\[X́AUSBzXg֘Âlrcs^XNCut@CłB

    2019-02-20 ޒH[(ނ) Rg̓{ꉻ(Japanese commentary)
 */
//DOM-IGNORE-END


#include <stdlib.h>
#include <string.h>
#include "usb.h"
#include "usb_host_msd.h"


// *****************************************************************************
// *****************************************************************************
// Section: Configuration
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
/* Max Number of Supported Devices

̒ĺÃNXhCoT|[głőڑfoCX\܂B
[U[l`Ȃꍇ́A1ɐݒ肳܂B
݁ÁAUSBzXgw̐ɂA1ɐݒ肷Kv܂B
*/
#ifndef USB_MAX_MASS_STORAGE_DEVICES
    #define USB_MAX_MASS_STORAGE_DEVICES        1
#endif

// *****************************************************************************
// *****************************************************************************
// Section: Constants
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// Section: State Machine Constants
// *****************************************************************************

#ifndef USB_ENABLE_TRANSFER_EVENT

    #define STATE_MASK                          0x00F0      //
    #define SUBSTATE_MASK                       0x000F      //

    #define NEXT_STATE                          0x0010      //
    #define NEXT_SUBSTATE                       0x0001      //


    #define STATE_DETACHED                      0x0000      //

    #define STATE_INITIALIZE_DEVICE             0x0010      //
    #define SUBSTATE_WAIT_FOR_ENUMERATION       0x0000      //
    #define SUBSTATE_SEND_GET_MAX_LUN           0x0001      //
    #define SUBSTATE_WAIT_FOR_MAX_LUN           0x0002      //
    #define SUBSTATE_GET_MAX_LUN_COMPLETE       0x0003      //

    #define STATE_RUNNING                       0x0020      //
    #define SUBSTATE_HOLDING                    0x0000      //
    #define SUBSTATE_SEND_CBW                   0x0001      //
    #define SUBSTATE_CBW_WAIT                   0x0002      //
    #define SUBSTATE_TRANSFER_DATA              0x0003      //
    #define SUBSTATE_TRANSFER_WAIT              0x0004      //
    #define SUBSTATE_REQUEST_CSW                0x0005      //
    #define SUBSTATE_CSW_WAIT                   0x0006      //
    #define SUBSTATE_TRANSFER_DONE              0x0007      //

    #define STATE_MSD_RESET_RECOVERY            0x0030      //
    #define SUBSTATE_SEND_RESET                 0x0000      //
    #define SUBSTATE_WAIT_FOR_RESET             0x0001      //
    #define SUBSTATE_RESET_COMPLETE             0x0002      //

    #define STATE_MSD_CLEAR_DATA_IN             0x0040      //
    #define SUBSTATE_SEND_CLEAR_IN              0x0000      //
    #define SUBSTATE_WAIT_FOR_CLEAR_IN          0x0001      //
    #define SUBSTATE_CLEAR_IN_COMPLETE          0x0002      //

    #define STATE_MSD_CLEAR_DATA_OUT            0x0050      //
    #define SUBSTATE_SEND_CLEAR_OUT             0x0000      //
    #define SUBSTATE_WAIT_FOR_CLEAR_OUT         0x0001      //
    #define SUBSTATE_CLEAR_OUT_COMPLETE         0x0002      //

    #define STATE_HOLDING                       0x0060      // G[ɂۗ

#else

    #define STATE_DETACHED                      0x0000      //

    #define STATE_WAIT_FOR_MAX_LUN              0x0001      //

    #define STATE_RUNNING                       0x0002      //
    #define STATE_CBW_WAIT                      0x0003      //
    #define STATE_TRANSFER_WAIT                 0x0004      //
    #define STATE_CSW_WAIT                      0x0005      //

    #define STATE_WAIT_FOR_RESET                0x0006      //
    #define STATE_WAIT_FOR_CLEAR_IN             0x0007      //
    #define STATE_WAIT_FOR_CLEAR_OUT            0x0008      //

    #define STATE_HOLDING                       0x0009      //  G[ɂۗ

    #define STATE_REQUEST_CSW                   0x000A      //  Dummy state

#endif


// *****************************************************************************
// Section: Other Constants
// *****************************************************************************

#define CBW_SIZE                            31              // CBW(Command Block Wrapper)̃oCg
#define CSW_SIZE                            13              // CSW(Command Status Wrapper)̃oCg

#define CSW_RECEIVE_ATTEMPTS                100             // CSW𓾂ׂ̂̃gC̐B

#define USB_MSD_GET_MAX_LUN                 0xFE            // őLUN擾ׂ̃foCXvR[hB
#define USB_MSD_RESET                       0xFF            // foCXZbgׂ̃foCXvR[hB
#define USB_MSD_DCBWSIGNATURE               0x43425355ul    // CBẄׂ̎ʒlB
#define USB_MSD_DCSWSIGNATURE               0x53425355ul    // CSŴׂ̈̎ʒlB

#define MARK_RESET_RECOVERY                 0x0E            // USB_MSD_DEVICE_INFOōXV܂B


//******************************************************************************
//******************************************************************************
// Section: Data Structures
//******************************************************************************
//******************************************************************************

// *****************************************************************************
// *****************************************************************************
/* CBW:Command Block Wrapper

́Ae}XXg[W]̊JnɕKvȁACommand Block Wrapper̍\łB
*/
typedef struct __attribute__((packed)) _USB_MSD_CBW
{
    uint32_t       dCBWSignature;              // CBWʂׂ̒萔
    uint32_t       dCBWTag;                    // zXgŔCӂɔԍ`^OBR}hƃXe[^XΉ܂B 
    uint32_t       dCBWDataTransferLength;     // f[^tF[Yœ]f[^̒w肵܂Bf[^ꍇ 0 łB 
    union
    {
        struct
        {
            uint8_t                : 7;
            uint8_t    bfDirection : 1;        // ] (0=OUT, 1=IN).
        };
        uint8_t        val;
    }           bmCBWflags;
    uint8_t        bCBWLUN;                    // 1USBfoCXɕ̃hCuڑĂꍇÃhCuԍw肵܂B(bits 3:0 only).
    uint8_t        bCBWCBLength;               // R}hpPbg̒B(bits 4:0 only).
    uint8_t        CBWCB[16];                  // R}hpPbgf[^
} USB_MSD_CBW;

// *****************************************************************************
/* CSW:Command Status Wrapper

́Ae}XXg[W]̍ŌɕKvȁACommand Status Wrapper̍\łB
*/
typedef struct __attribute__((packed)) _USB_MSD_CSW
{
    uint32_t   dCSWSignature;                  // CSWʂׂ̒萔
    uint32_t   dCSWTag;                        // R}h]dCBWTagɑΉ鎖ŁAzXgtF[ŸvmF鎖o܂B 
    uint32_t   dCSWDataResidue;                // ]Ȃf[^oCgB(0ȊO̓foCX̉f[^ZƂɂȂ܂B )
    uint8_t    dCSWStatus;                     // vꂽ̌(00H: 01H:s 02H:tF[YG[))
} USB_MSD_CSW;


/* 
  USB}XXg[WfoCXB
  ̍\́AڑĂeʋLuɊւSĂ̏ێׂɎgp܂B
*/
typedef struct _USB_MSD_DEVICE_INFO
{
    union __attribute__((packed)) {
        uint8_t                            data[31];               // foCXʐMp̃f[^obt@
        USB_MSD_CBW                     cbw;
        USB_MSD_CSW                     csw;
    }block;

    uint8_t                                deviceAddress;          // oX̃foCX̃AhX
    uint8_t                                clientDriverID;         // foCXvp̃NCAghCoID
    uint8_t                                errorCode;              // Ō̃G[̃G[R[h
    uint8_t                                state;                  // foCX̃Xe[g}V̏
    uint8_t                                returnState;            // G[sɖ߂
    union
    {
        struct
        {
            uint8_t                        bfDirection     : 1;    // ݂̓] (0=OUT, 1=IN).
            uint8_t                        bfReset         : 1;    // }XXg[WZbgs鎖tO
            uint8_t                        bfClearDataIN   : 1;    // INGh|CgNA鎖tO
            uint8_t                        bfClearDataOUT  : 1;    // OUTGh|CgNA鎖tO
        };
        uint8_t                            val;
    }                                   flags;
    uint8_t                                maxLUN;                 // foCX̍ő_jbgԍ
    uint8_t                                interface;              // gpĂC^[tFCXԍ
    uint8_t                                endpointIN;             // Bulk IN endpoint.
    uint8_t                                endpointOUT;            // Bulk OUT endpoint.
    uint8_t                                endpointDATA;           // ݂̓]ɎgpGh|Cg
    uint8_t                                *userData;              // [U[̃f[^obt@ւ̃|C^
    uint32_t                               userDataLength;         // [U[̃f[^obt@̒
    uint32_t                               bytesTransferred;       // [U[̃f[^obt@Ƃ̊Ԃœ]ꂽoCg
    uint32_t                               dCBWTag;                // dCSWtagƏƍׂdCBWTag̒l
    uint8_t                                attemptsCSW;            // CSW擾悤Ƃ񐔁B
} USB_MSD_DEVICE_INFO;

//******************************************************************************
//******************************************************************************
// Section: Local Prototypes
//******************************************************************************
//******************************************************************************

uint32_t   _USBHostMSD_GetNextTag( void );
void    _USBHostMSD_ResetStateJump( uint8_t i );


//******************************************************************************
//******************************************************************************
// Section: Macros
//******************************************************************************
//******************************************************************************

#ifndef USB_ENABLE_TRANSFER_EVENT
  #define _USBHostMSD_SetNextState()                { deviceInfoMSD[i].state = (deviceInfoMSD[i].state & STATE_MASK) + NEXT_STATE; }
  #define _USBHostMSD_SetNextSubState()             { deviceInfoMSD[i].state += NEXT_SUBSTATE; }
  #define _USBHostMSD_TerminateTransfer( error )    {                                                                           \
                                                        deviceInfoMSD[i].errorCode  = error;                                    \
                                                        deviceInfoMSD[i].state      = STATE_RUNNING | SUBSTATE_TRANSFER_DONE;   \
                                                    }
#else
  #ifdef USB_MSD_ENABLE_TRANSFER_EVENT
    #define _USBHostMSD_TerminateTransfer( error )  {                                                                                                           \
                                                        deviceInfoMSD[i].errorCode  = error;                                                                    \
                                                        deviceInfoMSD[i].state      = STATE_RUNNING;                                         \
                                                        usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_TRANSFER, NULL, 0 );     \
                                                    }
  #else
    #define _USBHostMSD_TerminateTransfer( error )  {                                                                                                           \
                                                        deviceInfoMSD[i].errorCode  = error;                                                                    \
                                                        deviceInfoMSD[i].state      = STATE_RUNNING;                                         \
                                                    }
  #endif
#endif


//******************************************************************************
//******************************************************************************
// Section: MSD Host External Variables
//******************************************************************************
//******************************************************************************
// ̃e[u̓AvP[ṼfBAC^[tF[Xẅׂ̏[`ƃCxgnh܂݂܂B
// USBConfig.cŒ`Ă܂B
extern CLIENT_DRIVER_TABLE usbMediaInterfaceTable;

//******************************************************************************
//******************************************************************************
// Section: MSD Host Global Variables
//******************************************************************************
//******************************************************************************

static uint32_t                       dCBWTagNext     = 0x12345678ul;
static USB_MSD_DEVICE_INFO         deviceInfoMSD[USB_MAX_MASS_STORAGE_DEVICES] __attribute__ ((aligned));


// *****************************************************************************
// *****************************************************************************
// Section: Application Callable Functions
// *****************************************************************************
// *****************************************************************************

/****************************************************************************
  Function:
    uint8_t USBHostMSDDeviceStatus( uint8_t deviceAddress )

  :
    ̋@\́AeʋLȕԂ肵܂B

  Precondition:
    None

  :
    uint8_t deviceAddress - ₢킹foCX̃AhX

  ߂l:
    USB_MSD_DEVICE_NOT_FOUND -  ȃfoCXAhXA܂̓foCXMSDł͂Ȃ
    USB_MSD_INITIALIZING     -  MSDڑĂAłB
    USB_MSD_NORMAL_RUNNING   -  MSD͒ʏ̎s[hł
    USB_MSD_RESETTING_DEVICE -  MSDZbg
    USB_MSD_DEVICE_DETACHED  -  MSD؂藣܂B(Ȃ͂)
    USB_MSD_ERROR_STATE      -  G[ׁ̈AMSDۗł(ʐM͂ł܂)

    ̑                   -  foCXGj[V̏ꍇ́A
 *                              USBHostDeviceStatus()̖߂R[hԂ܂B

  Remarks:
    None
  ***************************************************************************/

uint8_t USBHostMSDDeviceStatus( uint8_t deviceAddress )
{
    uint8_t    i;
    uint8_t    status;

    // LȃfoCXvĂ鎖mF܂B
    if ((deviceAddress == 0) || (deviceAddress > 127))
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    // foCX܂B
    for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_MASS_STORAGE_DEVICES)
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    // foCX݂̌̃Xe[^X𓾂
    status = USBHostDeviceStatus( deviceAddress );
    if (status != USB_DEVICE_ATTACHED)
    {
        return status;
    }
    else
    {
        // foCXڑăGj[V܂B
        // ̓IȏԂ擾鎖ł܂B

        #ifndef USB_ENABLE_TRANSFER_EVENT
            switch (deviceInfoMSD[i].state & STATE_MASK)
            {
                case STATE_INITIALIZE_DEVICE:
                    return USB_MSD_INITIALIZING;
                    break;

                case STATE_RUNNING:
                    return USB_MSD_NORMAL_RUNNING;
                    break;

                case STATE_HOLDING:
                    return USB_MSD_ERROR_STATE;
                    break;

                case STATE_MSD_RESET_RECOVERY:
                case STATE_MSD_CLEAR_DATA_IN:
                case STATE_MSD_CLEAR_DATA_OUT:
                    return USB_MSD_RESETTING_DEVICE;
                    break;

                default:
                    return USB_MSD_DEVICE_DETACHED;
                    break;
            }
        #else
            switch (deviceInfoMSD[i].state)
            {
                case STATE_WAIT_FOR_MAX_LUN:
                    return USB_MSD_INITIALIZING;
                    break;

                case STATE_RUNNING:
                case STATE_CBW_WAIT:
                case STATE_TRANSFER_WAIT:
                case STATE_CSW_WAIT:
                case STATE_REQUEST_CSW:
                    return USB_MSD_NORMAL_RUNNING;
                    break;

                case STATE_HOLDING:
                    return USB_MSD_ERROR_STATE;
                    break;

                case STATE_WAIT_FOR_RESET:
                case STATE_WAIT_FOR_CLEAR_IN:
                case STATE_WAIT_FOR_CLEAR_OUT:
                    return USB_MSD_RESETTING_DEVICE;
                    break;

                default:
                    return USB_MSD_DEVICE_DETACHED;
                    break;
            }
        #endif
    }
}


/****************************************************************************
  Function:
    uint8_t USBHostMSDResetDevice( uint8_t deviceAddress )

  Tv:
    ̋@\́AoNp̃}XXg[WZbgJn܂B

  :
    ZbǵAfoCXڑAĂȂꍇɂ̂ݔs鎖ł܂B

  Precondition:
    None

  :
    uint8_t deviceAddress - foCXAhX

  ߂l:
    USB_SUCCESS                 - ZbgJn
    USB_MSD_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_MSD_ILLEGAL_REQUEST     - foCX̓Zbgɑ΂ĕsȏԂł

  Remarks:
    None
  ***************************************************************************/

uint8_t USBHostMSDResetDevice( uint8_t deviceAddress )
{
    uint8_t    i;

    // LȑuvĂ鎖mF܂B
    if ((deviceAddress == 0) || (deviceAddress > 127))
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    // foCX܂B
    for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_MASS_STORAGE_DEVICES)
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    #ifndef USB_ENABLE_TRANSFER_EVENT
        if (((deviceInfoMSD[i].state & STATE_MASK) != STATE_DETACHED) &&
            ((deviceInfoMSD[i].state & STATE_MASK) != STATE_INITIALIZE_DEVICE))
    #else
        if ((deviceInfoMSD[i].state == STATE_RUNNING) ||
            (deviceInfoMSD[i].state == STATE_HOLDING ))
    #endif
    {
        deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].returnState = STATE_RUNNING | SUBSTATE_HOLDING;
        #else
            deviceInfoMSD[i].returnState = STATE_RUNNING;
        #endif
        _USBHostMSD_ResetStateJump( i );
        return USB_SUCCESS;
    }
    return USB_MSD_ILLEGAL_REQUEST;
}


/****************************************************************************
  Function:
    void USBHostMSDTasks( void )

  Tv:
    ̋@\́A}XXg[WNXɕKvȃeiX^XNs܂B

  :
    zXgw̓]CxggpĂȂꍇ́AAvP[VɂĒIɌĂяoKv܂B
    zXgw̓]CxggpĂꍇÅ֐̓RpCĂׁAĂяoKv͂܂B

  O:
    USBHostMSDInitialize()ĂяoĂ鎖B

  Parameters:
    None - None

  Returns:
    None

  Remarks:
    None
  ***************************************************************************/

void USBHostMSDTasks( void )
{

#ifndef USB_ENABLE_TRANSFER_EVENT

    uint32_t   byteCount;
    uint8_t    errorCode;
    uint8_t    i;

    for (i=0; i<USB_MAX_MASS_STORAGE_DEVICES; i++)
    {
        if (deviceInfoMSD[i].deviceAddress != 0)
        {
            switch (deviceInfoMSD[i].state & STATE_MASK)
            {
                case STATE_DETACHED:
                    // foCXڑĂ܂B
                    break;

                case STATE_INITIALIZE_DEVICE:
                    switch (deviceInfoMSD[i].state & SUBSTATE_MASK)
                    {
                        case SUBSTATE_WAIT_FOR_ENUMERATION:
                            if (USBHostDeviceStatus( deviceInfoMSD[i].deviceAddress ) == USB_DEVICE_ATTACHED)
                            {
                                _USBHostMSD_SetNextSubState();
                                #ifdef DEBUG_MODE
                                    // őLUN擾Ă܂...
                                    UART2PrintString( "MSD: Getting max LUN...\r\n" );
                                #endif
                            }
                            break;

                        case SUBSTATE_SEND_GET_MAX_LUN:
                            // ݃g[N𑗐MĂꍇ́Ał܂B
                            if (U1CONbits.TOKBUSY)
                                break;

                            if (!USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                                    USB_MSD_GET_MAX_LUN, 0, deviceInfoMSD[i].interface, 1, deviceInfoMSD[i].block.data, USB_DEVICE_REQUEST_GET, deviceInfoMSD[i].clientDriverID ))
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            else
                            {
                                // 蒼Ă݂܂B
                            }
                            break;

                        case SUBSTATE_WAIT_FOR_MAX_LUN:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                deviceInfoMSD[i].maxLUN = 0;
                                if (!errorCode)
                                {
                                    deviceInfoMSD[i].maxLUN = deviceInfoMSD[i].block.data[0];
                                }
                                else
                                {
                                    // STALLNAĂBEP0̏ꍇ́AXg[NAKv͂܂B
                                    USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, 0 );
                                }
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_GET_MAX_LUN_COMPLETE:
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Max LUN is " );
                                UART2PutHex( deviceInfoMSD[i].maxLUN );
                                UART2PrintString( "\r\nMSD: Running...\r\n" );
                            #endif

                            // MSDڑĂ鎖fBAC^[tFCXwɓ`܂B
                            if (usbMediaInterfaceTable.Initialize( deviceInfoMSD[i].deviceAddress, usbMediaInterfaceTable.flags, 0 ))
                            {
                                usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_MAX_LUN, &(deviceInfoMSD[i].maxLUN), 1 );
                                _USBHostMSD_SetNextState();
                            }
                            else
                            {
                                // fBAC^[tFCXw̓foCXT|[gł܂B
                                deviceInfoMSD[i].errorCode = USB_MSD_MEDIA_INTERFACE_ERROR;
                                deviceInfoMSD[i].state     = STATE_HOLDING;
                            }
                            break;
                    }
                    break;

                case STATE_RUNNING:
                    switch (deviceInfoMSD[i].state & SUBSTATE_MASK)
                    {
                        case SUBSTATE_HOLDING:
                            break;

                        case SUBSTATE_SEND_CBW:
                            // SCSIR}h̓]Jn
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Writing CBW\r\n" );
                            #endif
                            errorCode = USBHostWrite( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointOUT, deviceInfoMSD[i].block.data, CBW_SIZE );
                            if (errorCode)
                            {
                                _USBHostMSD_TerminateTransfer( errorCode );
                            }
                            else
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_CBW_WAIT:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointOUT, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    #ifdef DEBUG_MODE
                                        UART2PrintString( "MSD: Error with sending CBW\r\n" );
                                    #endif
                                    _USBHostMSD_TerminateTransfer( errorCode );
                                }
                                else if (byteCount != CBW_SIZE)
                                {
                                    #ifdef DEBUG_MODE
                                        UART2PrintString( "MSD: CBW size not correct\r\n" );
                                    #endif
                                    _USBHostMSD_TerminateTransfer( USB_MSD_CBW_ERROR );
                                }
                                else
                                {
                                    if (deviceInfoMSD[i].block.cbw.dCBWDataTransferLength == 0)
                                    {
                                        #ifdef DEBUG_MODE
                                            UART2PrintString( "MSD: Transfer length=0\r\n" );
                                        #endif
                                        // CSW肷ׂɃXLbv
                                        deviceInfoMSD[i].state = STATE_RUNNING | SUBSTATE_REQUEST_CSW;
                                    }
                                    else
                                    {
                                        #ifdef DEBUG_MODE
                                            UART2PrintString( "MSD: Going on...\r\n" );
                                        #endif
                                        _USBHostMSD_SetNextSubState();
                                    }
                                }
                            }
                            break;

                        case SUBSTATE_TRANSFER_DATA:
                            // f[^̓]tF[Y
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Transferring data, length ");
                                UART2PutHexDWord( deviceInfoMSD[i].userDataLength );
                                UART2PrintString( "\r\n" );
                            #endif
                            if (deviceInfoMSD[i].userDataLength == 0)
                            {
                                deviceInfoMSD[i].state = STATE_RUNNING | SUBSTATE_REQUEST_CSW;
                            }
                            else
                            {
                                if (!deviceInfoMSD[i].flags.bfDirection) // OUTpPbg(f[^̑M)
                                {
                                    errorCode = USBHostWrite( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA, deviceInfoMSD[i].userData, deviceInfoMSD[i].userDataLength );
                                }
                                else
                                {                                        // INpPbg(f[^̎M)
                                    errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA, deviceInfoMSD[i].userData, deviceInfoMSD[i].userDataLength );
                                }

                                if (errorCode)
                                {
                                    _USBHostMSD_TerminateTransfer( errorCode );
                                }
                                else
                                {
                                    _USBHostMSD_SetNextSubState();
                                }
                            }
                            break;

                        case SUBSTATE_TRANSFER_WAIT:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    if (errorCode == USB_ENDPOINT_STALLED)
                                    {
                                        // Xg[NAĂACSW肵ĂB
                                        #ifdef DEBUG_MODE
                                            UART2PrintString( "MSD: Stall on data\r\n" );
                                        #endif
                                        USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA );
                                        if (!deviceInfoMSD[i].flags.bfDirection) // OUT
                                        {
                                            deviceInfoMSD[i].flags.bfClearDataOUT = 1;
                                        }
                                        else
                                        {
                                            deviceInfoMSD[i].flags.bfClearDataIN = 1;
                                        }
                                        deviceInfoMSD[i].returnState = STATE_RUNNING | SUBSTATE_REQUEST_CSW;
                                        _USBHostMSD_ResetStateJump( i );

                                    }
                                    else
                                    {
                                        // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                        // cOȎɁA̃hCu͂Ń^[AEh^CG[X܂B
                                        // _USBHostMSD_TerminateTransferierrorCodej;
                                        deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
                                        deviceInfoMSD[i].returnState = STATE_RUNNING | SUBSTATE_SEND_CBW;   // ]蒼ĂB
                                        _USBHostMSD_ResetStateJump( i );
                                    }
                                }
                                else
                                {
                                    _USBHostMSD_SetNextSubState();
                                }
                            }
                            break;

                        case SUBSTATE_REQUEST_CSW:
                            // Xe[^X]tF[Y
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Getting CSW\r\n" );
                            #endif
                            // CSW̑MsXe[^X̓]vs
                            errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN, deviceInfoMSD[i].block.data, CSW_SIZE );
                            if (errorCode)
                            {
                                _USBHostMSD_TerminateTransfer( errorCode );
                            }
                            else
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_CSW_WAIT:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN, &errorCode, &byteCount ))
                            {
                                #ifdef DEBUG_MODE
                                    UART2PrintString( "MSD: Got CSW-" );
                                #endif
                                if (errorCode)
                                {
                                    deviceInfoMSD[i].attemptsCSW--;
                                    if (deviceInfoMSD[i].attemptsCSW)
                                    {
                                        USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN );
                                        deviceInfoMSD[i].flags.bfClearDataIN = 1;
                                        deviceInfoMSD[i].returnState = STATE_RUNNING | SUBSTATE_REQUEST_CSW;
                                        _USBHostMSD_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostMSD_TerminateTransfer( errorCode );
                                    }
                                }
                                else if ((byteCount != CSW_SIZE) |
                                         (deviceInfoMSD[i].block.csw.dCSWSignature != USB_MSD_DCSWSIGNATURE) |
                                         (deviceInfoMSD[i].block.csw.dCSWTag       != deviceInfoMSD[i].dCBWTag) )
                                {
                                    _USBHostMSD_TerminateTransfer( USB_MSD_CSW_ERROR );
                                }
                                else
                                {
                                    deviceInfoMSD[i].bytesTransferred = deviceInfoMSD[i].userDataLength - deviceInfoMSD[i].block.csw.dCSWDataResidue;

                                    if (deviceInfoMSD[i].block.csw.dCSWStatus != 0x00)
                                    {
                                        _USBHostMSD_TerminateTransfer( deviceInfoMSD[i].block.csw.dCSWStatus | USB_MSD_ERROR );
                                    }
                                    else
                                    {
                                        _USBHostMSD_TerminateTransfer( USB_SUCCESS );
                                    }

                                    // tF[YG[ꍇ́Aʏ̉^]ɖ߂̂ł͂ȂA[uuKv܂B
                                    if (deviceInfoMSD[i].block.csw.dCSWStatus == MSD_PHASE_ERROR)
                                    {
                                        deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
                                        deviceInfoMSD[i].returnState = STATE_RUNNING | SUBSTATE_HOLDING;
                                        _USBHostMSD_ResetStateJump( i );
                                    }
                                }
                            }
                            break;

                        case SUBSTATE_TRANSFER_DONE:
                            deviceInfoMSD[i].state = STATE_RUNNING | SUBSTATE_HOLDING;
                            #ifdef USB_MSD_ENABLE_TRANSFER_EVENT
                                usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_TRANSFER, NULL, 0 );
                            #endif
                            break;
                    }
                    break;

                case STATE_MSD_RESET_RECOVERY:
                    switch (deviceInfoMSD[i].state & SUBSTATE_MASK)
                    {
                        case SUBSTATE_SEND_RESET:
                            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                                            USB_MSD_RESET, 0, deviceInfoMSD[i].interface, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

                            if (errorCode)
                            {
                                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                            }
                            else
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_WAIT_FOR_RESET:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    //@ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                                }
                                else
                                {
                                    deviceInfoMSD[i].flags.bfReset = 0;
                                    _USBHostMSD_SetNextSubState();
                                }
                            }
                            break;

                        case SUBSTATE_RESET_COMPLETE:
                            _USBHostMSD_ResetStateJump( i );
                            break;
                    }
                    break;

                case STATE_MSD_CLEAR_DATA_IN:
                    switch (deviceInfoMSD[i].state & SUBSTATE_MASK)
                    {
                        case SUBSTATE_SEND_CLEAR_IN:
                            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_STANDARD | USB_SETUP_RECIPIENT_ENDPOINT,
                                            USB_REQUEST_CLEAR_FEATURE, USB_FEATURE_ENDPOINT_HALT, deviceInfoMSD[i].endpointIN, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

                            if (errorCode)
                            {
                                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                            }
                            else
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_WAIT_FOR_CLEAR_IN:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                                }
                                else
                                {
                                    deviceInfoMSD[i].flags.bfClearDataIN = 0;
                                    _USBHostMSD_SetNextSubState();
                                }
                            }
                            break;

                        case SUBSTATE_CLEAR_IN_COMPLETE:
                            _USBHostMSD_ResetStateJump( i );
                            break;
                    }
                    break;

                case STATE_MSD_CLEAR_DATA_OUT:
                    switch (deviceInfoMSD[i].state & SUBSTATE_MASK)
                    {
                        case SUBSTATE_SEND_CLEAR_OUT:
                            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_STANDARD | USB_SETUP_RECIPIENT_ENDPOINT,
                                            USB_REQUEST_CLEAR_FEATURE, USB_FEATURE_ENDPOINT_HALT, deviceInfoMSD[i].endpointOUT, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

                            if (errorCode)
                            {
                                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                            }
                            else
                            {
                                _USBHostMSD_SetNextSubState();
                            }
                            break;

                        case SUBSTATE_WAIT_FOR_CLEAR_OUT:
                            if (USBHostTransferIsComplete( deviceInfoMSD[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                                }
                                else
                                {
                                    deviceInfoMSD[i].flags.bfClearDataOUT = 0;
                                    _USBHostMSD_SetNextSubState();
                                }
                            }
                            break;

                        case SUBSTATE_CLEAR_OUT_COMPLETE:
                            _USBHostMSD_ResetStateJump( i );
                            break;
                    }
                    break;

                case STATE_HOLDING:
                    break;
            }
        }
    }
#endif
}


/****************************************************************************
  Function:
    void USBHostMSDTerminateTransfer( uint8_t deviceAddress )

  :
    ̋@\̓}XXg[W]I܂B

  Precondition:
    None

  :
    uint8_t deviceAddress  -foCXAhX

  Returns:
    None

  l:
    ̋@\sAfoCXɓ삵ׂɁA
 *  AvP[V̓foCXZbgKv܂B
  ***************************************************************************/

void USBHostMSDTerminateTransfer( uint8_t deviceAddress )
{
    uint8_t    i;

    // foCX܂B
    for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_MASS_STORAGE_DEVICES)
    {
        return;
    }

    // STATE_RUNNINGɂꍇɂ̂݁A]Ił܂B
    // łȂ΁A͑傫ȖNł傤B

    #ifndef USB_ENABLE_TRANSFER_EVENT
        if ((deviceInfoMSD[i].state & STATE_MASK) == STATE_RUNNING)
    #else
        if ((deviceInfoMSD[i].state == STATE_RUNNING) ||
            (deviceInfoMSD[i].state == STATE_CBW_WAIT) ||
            (deviceInfoMSD[i].state == STATE_TRANSFER_WAIT) ||
            (deviceInfoMSD[i].state == STATE_CSW_WAIT))
    #endif
    {
        // ĂGh|Cg]I܂B
        USBHostTerminateTransfer( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN );
        USBHostTerminateTransfer( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointOUT );

        // Ԃsɖ߂A]v҂Ă܂B
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].state = STATE_RUNNING | SUBSTATE_HOLDING;
        #else
            deviceInfoMSD[i].state = STATE_RUNNING;
        #endif
    }
    return;
}

/****************************************************************************
  Function:
    uint8_t USBHostMSDTransfer( uint8_t deviceAddress, uint8_t deviceLUN,
                uint8_t direction, uint8_t *commandBlock, uint8_t commandBlockLength,
                uint8_t *data, uint32_t dataLength )

  Tv:
   ̋@\̓}XXg[W]Jn܂B

  :
    ʏAAvP[V͂炭̋@\ɃANZXׂɓǂݎ/݃bp[𗘗p܂B

  Precondition:
    None

  :
    uint8_t deviceAddress      - foCXAhX
    uint8_t deviceLUN          - ANZXfoCXLUN
    uint8_t direction          - ]  1=read, 0=write
    uint8_t *commandBlock      - CBWp̃R}hubNւ̃|C^
    uint8_t commandBlockLength - R}hubN̒
    uint8_t *data              - f[^obt@ւ̃|C^
    uint32_t dataLength        - f[^obt@̃oCgTCY


  ߂l:
    USB_SUCCESS                 - NGXg͐ɊJn܂
    USB_MSD_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_MSD_DEVICE_BUSY         - ]sׂ̃foCXK؂ȏԂł͂܂
    USB_MSD_INVALID_LUN         - w肳ꂽLUN݂܂

  Remarks:
    None
  ***************************************************************************/

uint8_t USBHostMSDTransfer( uint8_t deviceAddress, uint8_t deviceLUN, uint8_t direction, uint8_t *commandBlock,
                        uint8_t commandBlockLength, uint8_t *data, uint32_t dataLength )
{
    uint8_t    i;
    uint8_t    j;

    #ifdef DEBUG_MODE
        UART2PrintString( "MSD: Transfer: " );
        if (direction)
        {
            UART2PrintString( "Read, " );
        }
        else
        {
            UART2PrintString( "Write, " );
        }
    #endif

    // LȑuvĂ鎖mF܂B
    if ((deviceAddress == 0) || (deviceAddress > 127))
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    // foCX܂B
    for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_MASS_STORAGE_DEVICES)
    {
        return USB_MSD_DEVICE_NOT_FOUND;
    }

    // foCXǂݏ\ȏԂɂȂĂ鎖mF܂B
    #ifndef USB_ENABLE_TRANSFER_EVENT
        if (deviceInfoMSD[i].state != (STATE_RUNNING | SUBSTATE_HOLDING))
    #else
        if (deviceInfoMSD[i].state != STATE_RUNNING)
    #endif
    {
        return USB_MSD_DEVICE_BUSY;
    }

    // ILUNmF܂B
    if (deviceLUN > deviceInfoMSD[i].maxLUN)
    {
        return USB_MSD_INVALID_LUN;
    }

    // ]܂B
    deviceInfoMSD[i].attemptsCSW       = CSW_RECEIVE_ATTEMPTS;
    deviceInfoMSD[i].bytesTransferred  = 0;
    deviceInfoMSD[i].errorCode         = USB_SUCCESS;
    deviceInfoMSD[i].flags.val         = 0;
    deviceInfoMSD[i].flags.bfDirection = direction;
    deviceInfoMSD[i].userData          = data;
    deviceInfoMSD[i].userDataLength    = dataLength;
    deviceInfoMSD[i].dCBWTag           = _USBHostMSD_GetNextTag();
    deviceInfoMSD[i].endpointDATA      = deviceInfoMSD[i].endpointIN;
    if (!direction) // OUTpPbg
    {
        deviceInfoMSD[i].endpointDATA  = deviceInfoMSD[i].endpointOUT;
    }
    #ifdef DEBUG_MODE
        UART2PrintString( "Data EP: " );
        UART2PutHex( deviceInfoMSD[i].endpointDATA );
        UART2PrintString( "\r\n" );
    #endif

    // [U[ɃR}hubNRAMԋpł悤ɁACBW܂B
    deviceInfoMSD[i].block.cbw.dCBWSignature             = USB_MSD_DCBWSIGNATURE;
    deviceInfoMSD[i].block.cbw.dCBWTag                   = deviceInfoMSD[i].dCBWTag;
    deviceInfoMSD[i].block.cbw.dCBWDataTransferLength    = deviceInfoMSD[i].userDataLength;
    deviceInfoMSD[i].block.cbw.bmCBWflags.val            = 0;
    deviceInfoMSD[i].block.cbw.bmCBWflags.bfDirection    = direction;
    deviceInfoMSD[i].block.cbw.bCBWLUN                   = deviceLUN;
    deviceInfoMSD[i].block.cbw.bCBWCBLength              = commandBlockLength;
    for (j=0; j<commandBlockLength; j++)
    {
        deviceInfoMSD[i].block.cbw.CBWCB[j]              = commandBlock[j];
    }

    #ifndef USB_ENABLE_TRANSFER_EVENT
        // ]ԂɃWv܂B
        deviceInfoMSD[i].state             = STATE_RUNNING | SUBSTATE_SEND_CBW;
    #else
        j = USBHostWrite( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointOUT, deviceInfoMSD[i].block.data, CBW_SIZE );
        if (j)
        {
            _USBHostMSD_TerminateTransfer( j );
        }
        else
        {
            deviceInfoMSD[i].state = STATE_CBW_WAIT;
        }
    #endif

    return USB_SUCCESS;
}


/****************************************************************************
  Function:
    bool USBHostMSDTransferIsComplete( uint8_t deviceAddress,
                        uint8_t *errorCode, uint32_t *byteCount )

  Tv:
    ̋@\́AŌ̓]ǂ܂B

  :
    ̋@\́AŌ̓]ǂ܂B
    ֐trueԂꍇAԂoCgƃG[R[h͗LłB
    xɎsł]1łAgpłGh|Cg1Ȃ̂ŁA
    foCXAhXm邾ōς݂܂B

  Precondition:
    None

  :
    uint8_t deviceAddress  - foCXAhX
    uint8_t *errorCode     - Ō̓]̃G[R[h
    uint32_t *byteCount    - ]oCg

  ߂l:
    true    - ]܂AerrorCode͗Lł
    false   - ]͊Ă܂AerrorCode͖ł

  Remarks:
    None
  ***************************************************************************/

bool USBHostMSDTransferIsComplete( uint8_t deviceAddress, uint8_t *errorCode, uint32_t *byteCount )
{
    uint8_t    i;

     // LȑuvĂ鎖mF܂B
    if ((deviceAddress == 0) || (deviceAddress > 127))
    {
        *errorCode = USB_MSD_DEVICE_NOT_FOUND;
        *byteCount = 0;
        return true;
    }

    // foCX܂B
    for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != deviceAddress); i++);
    if ((i == USB_MAX_MASS_STORAGE_DEVICES) || (deviceInfoMSD[i].state == STATE_DETACHED))
    {
        *errorCode = USB_MSD_DEVICE_NOT_FOUND;
        *byteCount = 0;
        return true;
    }

    #ifndef USB_ENABLE_TRANSFER_EVENT
        if  ( (deviceInfoMSD[i].state               == (STATE_RUNNING | SUBSTATE_HOLDING)) ||
             ((deviceInfoMSD[i].state & STATE_MASK) == STATE_HOLDING))
    #else
        if  ((deviceInfoMSD[i].state == STATE_RUNNING) ||
             (deviceInfoMSD[i].state == STATE_HOLDING))
    #endif
    {
        *byteCount = deviceInfoMSD[i].bytesTransferred;
        *errorCode = deviceInfoMSD[i].errorCode;

        return true;
    }
    else
    {
        return false;
    }
}


// *****************************************************************************
// *****************************************************************************
// Section: Host Stack Interface Functions
// *****************************************************************************
// *****************************************************************************

/****************************************************************************
  Function:
    bool USBHostMSDInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )

  Tv:
    ̊֐͂̃NCAghCȍ[`łB

  :
    USBfoCXGj[VĂ鎞ɃzXgwɂČĂяo܂B
    eʋLȕꍇ́AVfoCXp̃Xy[X鎖A
    yуfoCXɏȂƂ1̃oNINGh|Cg1̃oNOUTGh|Cg鎖mFKv܂B

  Precondition:
    None

  :
    uint8_t address        - VfoCX̃AhX
    uint32_t flags         - tO
    uint8_t clientDriverID - USBHostIssueDeviceRequest()AUSBHostSetDeviceConfiguration()A
                             USBHostSetDeviceInterface()ăfoCXNGXg𔭍s鎞ɑMID

 ߂l
    true   - foCXT|[g鎖ł܂B
    false  - foCXT|[g鎖͂ł܂B

  Remarks:
    None
  ***************************************************************************/

bool USBHostMSDInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )
{
    uint8_t   *descriptor;
    uint8_t    device;
    #ifdef USB_ENABLE_TRANSFER_EVENT
        uint8_t    errorCode;
    #endif
    uint16_t    i;
    uint8_t    endpointIN;
    uint8_t    endpointOUT;

    #ifdef DEBUG_MODE
        UART2PrintString( "MSD: USBHostMSDClientInitialize(0x" );
        UART2PutHex( flags );
        UART2PrintString( ")\r\n" );
    #endif

    // e[uŋ󂫃Xbg܂BȂꍇ́AfoCXLbNIt܂B
    for (device = 0; (device < USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[device].deviceAddress != 0); device++);
    if (device == USB_MAX_MASS_STORAGE_DEVICES)
    {
        #ifdef DEBUG_MODE
            // MSDp̋󂫃Xbg͂܂B
            UART2PrintString( "MSD: No free slots available for MSD.\r\n" );
        #endif
        // foCXLbNIt
        return false;
    }

    #ifdef DEBUG_MODE
        UART2PrintString( "MSD: MSD attached.\r\n" );
    #endif

    descriptor = USBHostGetCurrentConfigurationDescriptor( address );

    i = 0;

    #ifdef DEBUG_MODE
        UART2PrintString("MSD: Checking descriptor " );
        UART2PutDec( descriptor[i+5] );
        UART2PrintString(" ...\r\n" );
    #endif

    // ̃C^[tF[XfBXNv^܂B
    while (i < ((USB_CONFIGURATION_DESCRIPTOR *)descriptor)->wTotalLength)
    {
        #ifdef DEBUG_MODE
            UART2PrintString("MSD:  Checking interface...\r\n" );
        #endif
        // C^[tFCXfBXNv^wĂ邩ǂmF܂B
        if (descriptor[i+1] == USB_DESCRIPTOR_INTERFACE)
        {
            // C^[tF[XeʋLuC^[tF[Xł邩ǂmF܂B
            if (descriptor[i+5] == DEVICE_CLASS_MASS_STORAGE)
            {
                // C^[tFCX̃TuNXƃvgRǂmF܂B
                if (((descriptor[i+6] == DEVICE_SUBCLASS_SCSI) || (descriptor[i+6] == DEVICE_SUBCLASS_REMOVABLE)) &&
                    (descriptor[i+7] == DEVICE_INTERFACE_PROTOCOL_BULK_ONLY))
                {
                    deviceInfoMSD[device].interface = descriptor[i+2];

                    // oNINGh|CgOUTGh|CgT܂B
                    endpointIN  = 0;
                    endpointOUT = 0;

                    // Gh|CgfBXNv^XL܂B
                    i += descriptor[i];
                    while (descriptor[i+1] == USB_DESCRIPTOR_ENDPOINT)
                    {
                        if (descriptor[i+3] == 0x02) // Bulk
                        {
                            if (((descriptor[i+2] & 0x80) == 0x80) && (endpointIN == 0))
                            {
                                endpointIN = descriptor[i+2];
                            }
                            if (((descriptor[i+2] & 0x80) == 0x00) && (endpointOUT == 0))
                            {
                                endpointOUT = descriptor[i+2];
                            }
                        }
                        i += descriptor[i];
                    }

                    if ((endpointIN != 0) && (endpointOUT != 0))
                    {
                        // foCX܂B
                        deviceInfoMSD[device].deviceAddress    = address;
                        deviceInfoMSD[device].clientDriverID   = clientDriverID;
                        deviceInfoMSD[device].endpointIN       = endpointIN;
                        deviceInfoMSD[device].endpointOUT      = endpointOUT;
                        #ifdef DEBUG_MODE
                            UART2PrintString( "MSD: Bulk endpoint IN: " );
                            UART2PutHex( endpointIN );
                            UART2PrintString( " Bulk endpoint OUT: " );
                            UART2PutHex( endpointOUT );
                            UART2PrintString( "\r\n" );
                        #endif
                        USBHostSetNAKTimeout( address, endpointIN,  1, USB_NUM_BULK_NAKS );
                        USBHostSetNAKTimeout( address, endpointOUT, 1, USB_NUM_BULK_NAKS );

                        #ifndef USB_ENABLE_TRANSFER_EVENT
                            deviceInfoMSD[device].state = STATE_INITIALIZE_DEVICE;
                        #else
                            // őLUN擾ׂ̗vJn܂B 
                            // NGXgJnłȂꍇ́AfoCXGj[Vł܂B
                            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[device].deviceAddress,
                                    USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                                    USB_MSD_GET_MAX_LUN, 0, deviceInfoMSD[device].interface, 1, deviceInfoMSD[device].block.data,
                                    USB_DEVICE_REQUEST_GET, deviceInfoMSD[device].clientDriverID );
                            if (errorCode)
                            {
                                #ifdef DEBUG_MODE
                                    UART2PrintString( "MSD: Cannot get Max LUN - " );
                                    UART2PutHex( errorCode );
                                    UART2PrintString( "\r\n" );
                                #endif
                                return false;
                            }
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Getting Max LUN\r\n" );
                            #endif
                            deviceInfoMSD[device].state = STATE_WAIT_FOR_MAX_LUN;
                        #endif

                        return true;
                    }
                }
            }
        }

        // ̐ݒ̎̃fBXNv^ɃWv܂B
        i += descriptor[i];
    }

    // ̐ݒ͑eʋLuɂ͖łB
    return false;
}


/****************************************************************************
  Function:
    bool USBHostMSDEventHandler( uint8_t address, USB_EVENT event,
                            void *data, uint32_t size )

  Tv:
    ̊֐͂̃NCAghCõCxgnhłB

  :
    ܂܂ȃCxgɃzXgwɂČĂяo܂B

  O:
    foCXĂ鎖B

  :
    uint8_t address - foCX̃AhX
    USB_EVENT event - Cxg
    void *data      - CxgɊ֘Af[^ւ̃|C^
    uint16_t size   - f[^̃TCY

  ߂l:
    true   - Cxg܂
    false  - Cxg͏܂ł

  Remarks:
    None
  ***************************************************************************/

bool USBHostMSDEventHandler( uint8_t address, USB_EVENT event, void *data, uint32_t size )
{
    #if defined( USB_ENABLE_TRANSFER_EVENT )
        uint8_t    errorCode;
    #endif
    uint8_t    i;

    switch (event)
    {
        case EVENT_NONE:             // CxgĂȂ(NULL event)
            USBTasks();
            return true;
            break;

        case EVENT_DETACH:           // USBP[uOꂽ(data: uint8_t,foCX̃AhX)
            #ifdef DEBUG_MODE
                UART2PrintString( "MSD: Detach\r\n" );
            #endif

            // e[uŃfoCX܂B ꍇ́AdvȃtB[hNA܂B
            for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != address); i++);
            if (i < USB_MAX_MASS_STORAGE_DEVICES)
            {
                deviceInfoMSD[i].deviceAddress    = 0;
                deviceInfoMSD[i].state            = STATE_DETACHED;

                // ̏ʑwɃCxgʒm܂B
                usbMediaInterfaceTable.EventHandler( address, EVENT_DETACH, NULL, 0 );
            }
            return true;
            break;

        case EVENT_TRANSFER:         // USB]܂ - IvV
            #if defined( USB_ENABLE_TRANSFER_EVENT )
                #ifdef DEBUG_MODE
                    UART2PrintString( "MSD: transfer event: " );
                    UART2PutHex( address );
                    UART2PrintString( "\r\n" );
                #endif

                for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != address); i++) {}
                if (i == USB_MAX_MASS_STORAGE_DEVICES)
                {
                    #ifdef DEBUG_MODE
                        UART2PrintString( "MSD: Unknown device\r\n" );
                    #endif
                    return false;
                }

                #ifdef DEBUG_MODE
                    UART2PrintString( "MSD: Device state: " );
                    UART2PutHex( deviceInfoMSD[i].state );
                    UART2PrintString( "\r\n" );
                #endif
                switch (deviceInfoMSD[i].state)
                {
                    case STATE_WAIT_FOR_MAX_LUN:
                        #ifdef DEBUG_MODE
                            UART2PrintString( "MSD: Got Max LUN\r\n" );
                        #endif
                        deviceInfoMSD[i].maxLUN = 0;
                        if (!((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            deviceInfoMSD[i].maxLUN = deviceInfoMSD[i].block.data[0];
                        }
                        else
                        {
                            // Xg[NAĂB EP0̏ꍇ́AXg[NAKv͂܂B
                            USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, 0 );
                        }

                        // MSDڑĂ鎖fBAC^[tFCXwɓ`܂B
                        if (usbMediaInterfaceTable.Initialize( deviceInfoMSD[i].deviceAddress, usbMediaInterfaceTable.flags, 0 ))
                        {
                            usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_MAX_LUN, &(deviceInfoMSD[i].maxLUN), 1 );
                            deviceInfoMSD[i].state = STATE_RUNNING;
                        }
                        else
                        {
                            // fBAC^[tFCXw̓foCXT|[gł܂B
                            deviceInfoMSD[i].errorCode = USB_MSD_MEDIA_INTERFACE_ERROR;
                            deviceInfoMSD[i].state     = STATE_HOLDING;
                        }
                        return true;
                        break;

                    case STATE_RUNNING:
                        // œ]Cxg󂯎ׂł͂܂B
                        return false;
                        break;

                    case STATE_CBW_WAIT:
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Error with sending CBW\r\n" );
                            #endif
                            _USBHostMSD_TerminateTransfer( ((HOST_TRANSFER_DATA *)data)->bErrorCode );
                        }
                        else if (((HOST_TRANSFER_DATA *)data)->dataCount != CBW_SIZE)
                        {
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: CBW size not correct\r\n" );
                            #endif
                            _USBHostMSD_TerminateTransfer( USB_MSD_CBW_ERROR );
                        }
                        else
                        {
                            if ((((USB_MSD_CBW *)(deviceInfoMSD[i].block.data))->dCBWDataTransferLength == 0) ||
                                (deviceInfoMSD[i].userDataLength == 0))
                            {
                                #ifdef DEBUG_MODE
                                    UART2PrintString( "MSD: Transfer length=0\r\n" );
                                #endif

                                // CSW̓փXLbv
                                #ifdef DEBUG_MODE
                                    UART2PrintString( "MSD: Getting CSW\r\n" );
                                #endif

                                errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN, deviceInfoMSD[i].block.data, CSW_SIZE );
                                if (errorCode)
                                {
                                    _USBHostMSD_TerminateTransfer( errorCode );
                                }
                                else
                                {
                                    deviceInfoMSD[i].state = STATE_CSW_WAIT;
                                }
                            }
                            else
                            {
                                #ifdef DEBUG_MODE
                                    UART2PrintString( "MSD: Going on...\r\n" );
                                #endif
                                if (!deviceInfoMSD[i].flags.bfDirection) // OUT
                                {
                                    errorCode = USBHostWrite( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA, deviceInfoMSD[i].userData, deviceInfoMSD[i].userDataLength );
                                }
                                else
                                {
                                    errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA, deviceInfoMSD[i].userData, deviceInfoMSD[i].userDataLength );
                                }

                                if (errorCode)
                                {
                                    _USBHostMSD_TerminateTransfer( errorCode );
                                }
                                else
                                {
                                    deviceInfoMSD[i].state = STATE_TRANSFER_WAIT;
                                }
                            }
                        }
                        return true;
                        break;

                    case STATE_TRANSFER_WAIT:
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            if (((HOST_TRANSFER_DATA *)data)->bErrorCode == USB_ENDPOINT_STALLED)
                            {
                                // Xg[NAĂACSW肵ĂB
                                USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA );
                                if (!deviceInfoMSD[i].flags.bfDirection) // OUT
                                {
                                    deviceInfoMSD[i].flags.bfClearDataOUT = 1;
                                }
                                else
                                {
                                    deviceInfoMSD[i].flags.bfClearDataIN = 1;
                                }
                                deviceInfoMSD[i].returnState = STATE_REQUEST_CSW;
                                _USBHostMSD_ResetStateJump( i );

                            }
                            else
                            {
                                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                                // _USBHostMSD_TerminateTransfer( errorCode );
                                deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
                                deviceInfoMSD[i].returnState = STATE_RUNNING;
                                _USBHostMSD_ResetStateJump( i );
                            }
                        }
                        else
                        {
                            #ifdef DEBUG_MODE
                                UART2PrintString( "MSD: Getting CSW\r\n" );
                            #endif

                            errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN, deviceInfoMSD[i].block.data, CSW_SIZE );
                            if (errorCode)
                            {
                                _USBHostMSD_TerminateTransfer( errorCode );
                            }
                            else
                            {
                                deviceInfoMSD[i].state = STATE_CSW_WAIT;
                            }
                        }
                        break;

                    case STATE_CSW_WAIT:
                        #ifdef DEBUG_MODE
                            UART2PrintString( "MSD: Got CSW - " );
                        #endif
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            deviceInfoMSD[i].attemptsCSW--;
                            if (deviceInfoMSD[i].attemptsCSW)
                            {
                                USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN );
                                deviceInfoMSD[i].flags.bfClearDataIN = 1;
                                deviceInfoMSD[i].returnState = STATE_REQUEST_CSW;
                                _USBHostMSD_ResetStateJump( i );
                            }
                            else
                            {
                                _USBHostMSD_TerminateTransfer( ((HOST_TRANSFER_DATA *)data)->bErrorCode );
                            }
                        }
                        else if ((((HOST_TRANSFER_DATA *)data)->dataCount != CSW_SIZE) |
                                 (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWSignature != USB_MSD_DCSWSIGNATURE) |
                                 (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWTag       != deviceInfoMSD[i].dCBWTag) )
                        {
                            _USBHostMSD_TerminateTransfer( USB_MSD_CSW_ERROR );
                        }
                        else
                        {
                            deviceInfoMSD[i].bytesTransferred = deviceInfoMSD[i].userDataLength - ((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWDataResidue;

                            if (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWStatus != 0x00)
                            {
                                _USBHostMSD_TerminateTransfer( ((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWStatus | USB_MSD_ERROR );
                            }
                            else
                            {
                                _USBHostMSD_TerminateTransfer( USB_SUCCESS );
                            }

                            // tF[YG[ꍇ́Aʏ̉^]ɖ߂̂ł͂ȂA[uuKv܂B
                            if (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWStatus == MSD_PHASE_ERROR)
                            {
                                deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
                                deviceInfoMSD[i].returnState = STATE_RUNNING;
                                _USBHostMSD_ResetStateJump( i );
                            }
                        }
                        break;

                    case STATE_WAIT_FOR_RESET:
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                            _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                        }
                        else
                        {
                            deviceInfoMSD[i].flags.bfReset = 0;
                            _USBHostMSD_ResetStateJump( i );
                        }
                        break;

                    case STATE_WAIT_FOR_CLEAR_IN:
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                            _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                        }
                        else
                        {
                            deviceInfoMSD[i].flags.bfClearDataIN = 0;
                            _USBHostMSD_ResetStateJump( i );
                        }
                        break;

                    case STATE_WAIT_FOR_CLEAR_OUT:
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                        {
                            // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                            _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                        }
                        else
                        {
                            deviceInfoMSD[i].flags.bfClearDataOUT = 0;
                            _USBHostMSD_ResetStateJump( i );
                        }
                        break;

                    case STATE_HOLDING:
                        break;
                }
            #endif

        case EVENT_SOF:              // t[̎n܂ - Kv܂
        case EVENT_RESUME:           // MfoCXE[h̍ĊJ
        case EVENT_SUSPEND:          // foCX[h̃TXyh/AChCxgM
        case EVENT_RESET:            // foCX[h̃oXZbgM
        case EVENT_STALL:            // Xg[܂
            return true;
            break;

        case EVENT_BUS_ERROR:        // USBɓ]G[܂B
            #if defined( USB_ENABLE_TRANSFER_EVENT )
            #ifdef DEBUG_MODE
                UART2PrintString( "MSD: transfer event: " );
                UART2PutHex( address );
                UART2PrintString( "\r\n" );
            #endif

            for (i=0; (i<USB_MAX_MASS_STORAGE_DEVICES) && (deviceInfoMSD[i].deviceAddress != address); i++) {}
            if (i == USB_MAX_MASS_STORAGE_DEVICES)
            {
                #ifdef DEBUG_MODE
                    UART2PrintString( "MSD: Unknown device\r\n" );
                #endif
                return false;
            }

            #ifdef DEBUG_MODE
                UART2PrintString( "MSD: Device state: " );
                UART2PutHex( deviceInfoMSD[i].state );
                UART2PrintString( "\r\n" );
            #endif
            switch (deviceInfoMSD[i].state)
            {
                case STATE_WAIT_FOR_MAX_LUN:
                    #ifdef DEBUG_MODE
                        UART2PrintString( "MSD: Got Max LUN\r\n" );
                    #endif

                     // STALLNAĂB EP0̏ꍇ́AXg[NAKv͂܂B
                     USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, 0 );

                    // MSDڑĂ鎖fBAC^[tFCXwɓ`܂B
                    if (usbMediaInterfaceTable.Initialize( deviceInfoMSD[i].deviceAddress, usbMediaInterfaceTable.flags, 0 ))
                    {
                        usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_MAX_LUN, &(deviceInfoMSD[i].maxLUN), 1 );
                        deviceInfoMSD[i].state = STATE_RUNNING;
                    }
                    else
                    {
                        // fBAC^[tFCXw̓foCXT|[gł܂B
                        deviceInfoMSD[i].errorCode = USB_MSD_MEDIA_INTERFACE_ERROR;
                        deviceInfoMSD[i].state     = STATE_HOLDING;
                    }
                    return true;
                    break;

                case STATE_RUNNING:
                    // œ]Cxg󂯎ׂł͂܂B
                    return false;
                    break;

                case STATE_CBW_WAIT:
                    if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                    {
                        #ifdef DEBUG_MODE
                            UART2PrintString( "MSD: Error with sending CBW\r\n" );
                        #endif
                        _USBHostMSD_TerminateTransfer( ((HOST_TRANSFER_DATA *)data)->bErrorCode );
                    }
                    else if (((HOST_TRANSFER_DATA *)data)->dataCount != CBW_SIZE)
                    {
                        #ifdef DEBUG_MODE
                            UART2PrintString( "MSD: CBW size not correct\r\n" );
                        #endif
                        _USBHostMSD_TerminateTransfer( USB_MSD_CBW_ERROR );
                    }
                    return true;
                    break;

                case STATE_TRANSFER_WAIT:
                    if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                    {
                        if (((HOST_TRANSFER_DATA *)data)->bErrorCode == USB_ENDPOINT_STALLED)
                        {
                            // Xg[NAĂACSW肵ĂB
                            USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointDATA );
                            if (!deviceInfoMSD[i].flags.bfDirection) // OUT
                            {
                                deviceInfoMSD[i].flags.bfClearDataOUT = 1;
                            }
                            else
                            {
                                deviceInfoMSD[i].flags.bfClearDataIN = 1;
                            }
                            deviceInfoMSD[i].returnState = STATE_REQUEST_CSW;
                            _USBHostMSD_ResetStateJump( i );

                        }
                        else
                        {
                            // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                            // _USBHostMSD_TerminateTransfer( errorCode );
                            deviceInfoMSD[i].flags.val |= MARK_RESET_RECOVERY;
                            deviceInfoMSD[i].returnState = STATE_RUNNING;
                            _USBHostMSD_ResetStateJump( i );
                        }
                    }
                    break;

                case STATE_CSW_WAIT:
                    if (((HOST_TRANSFER_DATA *)data)->bErrorCode)
                    {
                        deviceInfoMSD[i].attemptsCSW--;
                        if (deviceInfoMSD[i].attemptsCSW)
                        {
                            USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN );
                            deviceInfoMSD[i].flags.bfClearDataIN = 1;
                            deviceInfoMSD[i].returnState = STATE_REQUEST_CSW;
                            _USBHostMSD_ResetStateJump( i );
                        }
                        else
                        {
                            _USBHostMSD_TerminateTransfer( ((HOST_TRANSFER_DATA *)data)->bErrorCode );
                        }
                    }
                    else if ((((HOST_TRANSFER_DATA *)data)->dataCount != CSW_SIZE) |
                             (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWSignature != USB_MSD_DCSWSIGNATURE) |
                             (((USB_MSD_CSW *)(deviceInfoMSD[i].block.data))->dCSWTag       != deviceInfoMSD[i].dCBWTag) )
                    {
                        _USBHostMSD_TerminateTransfer( USB_MSD_CSW_ERROR );
                    }
                    break;

                case STATE_WAIT_FOR_RESET:
                    // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                    break;

                case STATE_WAIT_FOR_CLEAR_IN:
                    // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                    break;

                case STATE_WAIT_FOR_CLEAR_OUT:
                    // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                    _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
                    break;

                case STATE_HOLDING:
                    break;
            }
            return true;
            #endif
        default:
            return false;
            break;
    }

    return false;
}


// *****************************************************************************
// *****************************************************************************
// Section: Internal Functions
// *****************************************************************************
// *****************************************************************************

/****************************************************************************
  Function:
    uint32_t _USBHostMSD_GetNextTag( void )

  :
    eoN]ɂ́ACBWɑ݂ACSWňvȂ΂ȂȂ]ID܂B
    ̊֐́A]p̎IDԍ擾܂B

  Precondition:
    None

  Parameters:
    None - None

  ߂l:
    uint32_t - CBWŎgp^O

  Remarks:
    None
  ***************************************************************************/

uint32_t _USBHostMSD_GetNextTag( void )
{
    return dCBWTagNext++;
}


/****************************************************************************
  Function:
    void _USBHostMSD_ResetStateJump( uint8_t i )

  Summary:


  :
    ̊֐́AɃZbĝǂ̕sKv邩𔻒fȀԂɃWv܂B

  Precondition:
    foCXdeviceInfoMSDzɂȂ΂Ȃ܂B

  Parameters:
    uint8_t i  - ZbgfoCXdeviceInfoMSD\̂ւ̃CfbNX

  Returns:
    None

  Remarks:
    None
  ***************************************************************************/

void _USBHostMSD_ResetStateJump( uint8_t i )
{
    #ifdef USB_ENABLE_TRANSFER_EVENT
        uint8_t    errorCode;
    #endif

    if (deviceInfoMSD[i].flags.bfReset)
    {
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].state = STATE_MSD_RESET_RECOVERY;
        #else
            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                            USB_MSD_RESET, 0, deviceInfoMSD[i].interface, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

            if (errorCode)
            {
                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
            }
            else
            {
                deviceInfoMSD[i].state = STATE_WAIT_FOR_RESET;
            }
        #endif
    }
    else if (deviceInfoMSD[i].flags.bfClearDataIN)
    {
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].state = STATE_MSD_CLEAR_DATA_IN;
        #else
            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_STANDARD | USB_SETUP_RECIPIENT_ENDPOINT,
                            USB_REQUEST_CLEAR_FEATURE, USB_FEATURE_ENDPOINT_HALT, deviceInfoMSD[i].endpointIN, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

            if (errorCode)
            {
                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
            }
            else
            {
                deviceInfoMSD[i].state = STATE_WAIT_FOR_CLEAR_IN;
            }
        #endif
    }
    else if (deviceInfoMSD[i].flags.bfClearDataOUT)
    {
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].state = STATE_MSD_CLEAR_DATA_OUT;
        #else
            errorCode = USBHostIssueDeviceRequest( deviceInfoMSD[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_STANDARD | USB_SETUP_RECIPIENT_ENDPOINT,
                            USB_REQUEST_CLEAR_FEATURE, USB_FEATURE_ENDPOINT_HALT, deviceInfoMSD[i].endpointOUT, 0, NULL, USB_DEVICE_REQUEST_SET, deviceInfoMSD[i].clientDriverID );

            if (errorCode)
            {
                // ł̃G[񕜂ɂẮAdlł͖IɐĂ܂B
                _USBHostMSD_TerminateTransfer( USB_MSD_RESET_ERROR );
            }
            else
            {
                deviceInfoMSD[i].state = STATE_WAIT_FOR_CLEAR_OUT;
            }
        #endif
    }
    else
    {
        usbMediaInterfaceTable.EventHandler( deviceInfoMSD[i].deviceAddress, EVENT_MSD_RESET, NULL, 0 );

        // G[NAāAx蒼ĂB
        USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN );
        USBHostClearEndpointErrors( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointOUT );

        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoMSD[i].state = deviceInfoMSD[i].returnState;
        #else
            if (deviceInfoMSD[i].returnState == STATE_REQUEST_CSW)
            {
                // CSWv
                #ifdef DEBUG_MODE
                    UART2PrintString( "MSD: Getting CSW\r\n" );
                #endif

                errorCode = USBHostRead( deviceInfoMSD[i].deviceAddress, deviceInfoMSD[i].endpointIN, deviceInfoMSD[i].block.data, CSW_SIZE );
                if (errorCode)
                {
                    _USBHostMSD_TerminateTransfer( errorCode );
                }
                else
                {
                    deviceInfoMSD[i].state = STATE_CSW_WAIT;
                }
            }
            else
            {
                deviceInfoMSD[i].state = deviceInfoMSD[i].returnState;
            }
        #endif

    }
}


