// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright 2015 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/
/*  ̃\[X́AXg[WfoCXSCSIŃC^[tFCXsvOłB

    2019-02-20 ޒH[(ނ) Rg̓{ꉻ(Japanese commentary)
                                   SCSIR}hs̃gC񐔂 5->100 ɕύX
 */
//DOM-IGNORE-END

#include <stdlib.h>
#include <string.h>
#include <usb_config.h>
#include <fileio.h>
#include <usb.h>
#include <usb_host_msd.h>
#include <usb_host_msd_scsi.h>

//#define DEBUG_MODE   // fobO\Lɂ܂
#if defined(DEBUG_MODE)
    #include "../skDEBUG.h"
#endif

#define PERFORM_TEST_UNIT_READY

// *****************************************************************************
// *****************************************************************************
// Section: Constants
// *****************************************************************************
// *****************************************************************************

#define FAT_GOOD_SIGN_0             0x55        // Lȃp[eBVSignature 0tB[h̒lB
#define FAT_GOOD_SIGN_1             0xAA        // Lȃp[eBVSignature 1tB[h̒lB
#define FO_MBR                      0L          // }X^[u[gR[hZN^[LBA
#define FUA_ALLOW_CACHE             0x00        // jbgANZXALbV̎gp
#define INITIALIZATION_ATTEMPTS     100         // s܂łɃfBȀs
#define RDPROTECT_NORMAL            0x00        // ʏ̓ǂݎی쓮B
#define WRPROTECT_NORMAL            0x00        // ʏ̏݋֎~B


//******************************************************************************
//******************************************************************************
// Section: Local Prototypes and Macros
//******************************************************************************
//******************************************************************************

#if !defined( USBTasks )
    #define USBTasks()                  \
        {                               \
            USBHostTasks();             \
            USBHostMSDTasks();          \
        }
#endif

#if defined( PERFORM_TEST_UNIT_READY )
    bool    _USBHostMSDSCSI_TestUnitReady( uint8_t * address );
#endif

static bool USBHostMSDSCSIRequestSense(uint8_t * address);

//******************************************************************************
//******************************************************************************
// Section: SCSI MSD Host Global Variables
//******************************************************************************
//******************************************************************************

//******************************************************************************
// Section: Internal Global Variables
//******************************************************************************

static FILEIO_MEDIA_INFORMATION   mediaInformation;   // t̃fBAɊւB

// *****************************************************************************
// *****************************************************************************
// Section: MSD Host Stack Callback Functions
// *****************************************************************************
// *****************************************************************************

/****************************************************************************
  Function:
    bool USBHostMSDSCSIInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )

  :
    ̋@\́AUSBeʋLuGj[VĂ鎞ɌĂяo܂B

  Precondition:
    None

  :
    uint8_t address        -   VfoCX̃AhX
    uint32_t flags         -   tO
    uint8_t clientDriverID -   ̑ẅׂIDAfBAC^tF[Xwł͎gp܂B

  ߂l:
    true    -  foCXT|[g鎖ł܂B
    false   -  foCXT|[g鎖͂ł܂B

  Remarks:
    None
  ***************************************************************************/

bool USBHostMSDSCSIInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )
{
    #ifdef DEBUG_MODE
        // foCXtĂ܂B
        UART2PrintString( "SCSI: Device attached.\r\n" );
    #endif
    return true;
}


/****************************************************************************
  Function:
    bool USBHostMSDSCSIEventHandler( uint8_t address, USB_EVENT event,
                        void *data, uint32_t size )

  :
    ̊֐́AUSBzXg}XXg[WNCAghCoł܂܂ȃCxgƂɌĂяo܂B

  O:
    foCXĂ鎖B

  :
    uint8_t address    -   foCX̃AhX
    USB_EVENT event    -   Cxg
    void *data         -   CxgɊ֘Af[^ւ̃|C^
    uint32_t size      -   Cxg̃f[^TCY

  ߂l:
    true    -   Cxg܂
    false   -   Cxg͏܂ł

  Remarks:
    None
  ***************************************************************************/

bool USBHostMSDSCSIEventHandler( uint8_t address, USB_EVENT event, void *data, uint32_t size )
{
    switch( (int) event )
    {
        case EVENT_MSD_NONE:       // Cxg͔ĂȂiNULLCxgj
        case EVENT_MSD_TRANSFER:   // MSD]܂
            return true;
            break;

        case EVENT_MSD_RESET:      // MSDZbg
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: MSD Reset performed.\r\n" );
            #endif
            return true;
            break;

        case EVENT_MSD_MAX_LUN:    // foCX̍őLUNݒ肵܂
            {
                uint8_t i;
                #ifdef DEBUG_MODE
                    UART2PrintString( "SCSI: Max LUN set.\r\n" );
                #endif
                mediaInformation.maxLUN                     = *((uint8_t *)data);
                mediaInformation.validityFlags.bits.maxLUN  = 1;

                for (i = 0; i < mediaInformation.maxLUN + 1; i++)
                {
                    // AvP[VMSDfoCXڑꂽ񍐂
                    USB_HOST_APP_EVENT_HANDLER (address, EVENT_MSD_ATTACH, &i, 1);
                }
            }
            return true;
            break;

        case EVENT_DETACH:         // USBP[uOꂽ(f[^Fuint8_t foCX̃AhX)
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: Device detached.\r\n" );
            #endif
            address                              = 0;
            mediaInformation.validityFlags.value = 0;
            return true;
            break;

        default:
            return false;
            break;
    }
    return false;
}


// *****************************************************************************
// *****************************************************************************
// Section: Application Callable Functions
// *****************************************************************************
// *****************************************************************************

/****************************************************************************
  Function:
    uint8_t USBHostMSDSCSIMediaDetect( uint8_t * address )

  :
    ̋@\́AeʋLuڑĎgp\ǂ𔻒f܂B

  Precondition:
    None

  :
    uint8_t *address    -   foCX̃AhXւ̃|C^

  ߂l:
    true    -   MSD݂AłĂ
    false   -   MSD݂ȂłĂ܂

  l:
    ̓foCX҂ԃ[vŌĂяo鎖̂ŁA
    USB^XNs鎖mFKv܂B
  ***************************************************************************/

uint8_t USBHostMSDSCSIMediaDetect( uint8_t * address)
{
    if (USBHostMSDDeviceStatus( *(uint8_t *)address ) == USB_MSD_NORMAL_RUNNING)
    {
        return true;
    }

    return false;
}


/****************************************************************************
  Function:
    MEDIA_INFORMATION * USBHostMSDSCSIMediaInitialize( uint8_t * address )

  :
    ̋@\̓fBA܂B

  Precondition:
    None

  :
    uint8_t *address    -   foCX̃AhXւ̃|C^

  ߂l:
    ̊֐MEDIA_INFORMATION\̂ւ̃|C^Ԃ܂B
    errorCodeo[ɂ́Aȉ̒l܂߂鎖ł܂B
        * MEDIA_NO_ERROR - fBA͐ɏAZN^TCY͗LɂȂ܂
                          ivalidityFlagsrbggpĊmFĂj
        * MEDIA_DEVICE_NOT_PRESENT - vꂽfoCX͐ڑĂ܂
        * MEDIA_CANNOT_INITIALIZE  - fBAł܂B

  l:
    ̊֐͎SCSIR}hs܂B
        * READ CAPACITY(10) - Xg[W̗eʏv܂B
        * REQUEST SENSE     - ŌɎsꂽR}h̃G[v܂B

    READ CAPACITY(10) R}hubN͎̂ƂłB

    <code>
        Byte/Bit    7       6       5       4       3       2       1       0
           0                    Operation Code (0x25)
           1        [                      Reserved                         ]
           2        [ (MSB)
           3                        Logical Block Address
           4
           5                                                          (LSB) ]
           6        [                      Reserved
           7                                                                ]
           8        [                      Reserved                 ] [ PMI ]
           9        [                    Control                            ]
    </code>

    REQUEST SENSE R}hubN͎̂ƂłB

    <code>
        Byte/Bit    7       6       5       4       3       2       1       0
           0                    Operation Code (0x02)
           1        [                      Reserved                 ] [ DESC]
           2        [                      Reserved
           3                                                                ]
           4        [                  Allocation Length                    ]
           5        [                    Control                            ]
    </code>
  ***************************************************************************/

FILEIO_MEDIA_INFORMATION * USBHostMSDSCSIMediaInitialize( uint8_t * address )
{
    uint8_t        attempts;
    uint32_t       byteCount;
    uint8_t        commandBlock[10];
    uint8_t        errorCode;
    uint8_t        inquiryData[36];

    // foCX܂ڑĂ鎖mF܂B
    if (*address == 0)
    {
        // vꂽfoCX݂͑܂
        mediaInformation.errorCode = MEDIA_DEVICE_NOT_PRESENT;
        return &mediaInformation;
    }

    attempts = INITIALIZATION_ATTEMPTS;
    while (attempts != 0)
    {
        attempts --;

        #ifdef DEBUG_MODE
            UART2PrintString( "SCSI: READ CAPACITY 10..." );
        #endif

        // R}hubN READ CAPACITY(10) p[^Zbg܂B
        // Xg[W̗eʏv܂B
        commandBlock[0] = 0x25;     // R[h
        commandBlock[1] = 0;        //
        commandBlock[2] = 0;        //
        commandBlock[3] = 0;        //
        commandBlock[4] = 0;        //
        commandBlock[5] = 0;        //
        commandBlock[6] = 0;        //
        commandBlock[7] = 0;        //
        commandBlock[8] = 0;        //
        commandBlock[9] = 0x00;     // Control

        // eʋLu̓ǂݎJn܂B
        errorCode = USBHostMSDRead( *address, 0, commandBlock, 10, inquiryData, 8 );
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PutChar( ' ' );
        #endif

        if (!errorCode)
        {   // ]܂ő҂܂B
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

        if (!errorCode)
        {   // 
            #ifdef DEBUG_MODE
                UART2PutHex( errorCode ) ;
                UART2PrintString( "\r\n" );
            #endif

            // ZN^[TCY肵܂B
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: Sector size:" );
                UART2PutChar( inquiryData[7] + '0' );
                UART2PutChar( inquiryData[6] + '0' );
                UART2PutChar( inquiryData[5] + '0' );
                UART2PutChar( inquiryData[4] + '0' );
                UART2PrintString( "\r\n" );
            #endif
            mediaInformation.sectorSize                     = (inquiryData[7] << 12) + (inquiryData[6] << 8) + (inquiryData[5] << 4) + (inquiryData[4]);
            mediaInformation.validityFlags.bits.sectorSize  = 1;

            mediaInformation.errorCode = MEDIA_NO_ERROR;
            return &mediaInformation;   // ʏ͂ŐRETURN
        }
        else
        {   // ŝŃR}h̃G[v܂B
            // Xg[NA悤ƂׂɁAREQUEST SENSE s܂B
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: REQUEST SENSE..." );
            #endif

            // R}hubNREQUEST SENSEp[^Zbg܂B
            commandBlock[0] = 0x03;     // R[h
            commandBlock[1] = 0;        //
            commandBlock[2] = 0;        //
            commandBlock[3] = 0;        //
            commandBlock[4] = 18;       // 蓖Ē
            commandBlock[5] = 0;        // Control

            errorCode = USBHostMSDRead( *address, 0, commandBlock, 6, inquiryData, 18 );
            #ifdef DEBUG_MODE
                UART2PutHex( errorCode ) ;
                UART2PutChar( ' ' );
            #endif

            if (!errorCode)
            {
                while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
                {
                    USBTasks();
                }
            }
        }
    }

    // "TEST UNIT READY"R}h𔭍săfoCX]ɑΉ\ǂ₢킹sĂ
    // ēxAXg[W̗eʏv܂B
    attempts = INITIALIZATION_ATTEMPTS;
    while (attempts != 0)
    {
        attempts --;

        // Xg[NA悤ƂׂɁAREQUEST SENSE s܂B
        #ifdef DEBUG_MODE
            UART2PrintString( "SCSI: REQUEST SENSE..." );
        #endif

        // R}hubNREQUEST SENSEp[^Zbg܂B
        commandBlock[0] = 0x03;     // R[h
        commandBlock[1] = 0;        //
        commandBlock[2] = 0;        //
        commandBlock[3] = 0;        //
        commandBlock[4] = 18;       // 蓖Ē
        commandBlock[5] = 0;        // Control

        errorCode = USBHostMSDRead( *address, 0, commandBlock, 6, inquiryData, 18 );
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PutChar( ' ' );
        #endif

        if (!errorCode)
        {
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

       // "TEST UNIT READY"R}h𔭍săfoCX]ɑΉ\ǂ₢킹܂B
        _USBHostMSDSCSI_TestUnitReady(address);


        // NA悤Ƃ邽߂ɁARequest Senses܂B
        #ifdef DEBUG_MODE
            UART2PrintString( "SCSI: REQUEST SENSE..." );
        #endif

        // R}hubNREQUEST SENSEp[^͂܂B
        commandBlock[0] = 0x03;     // R[h
        commandBlock[1] = 0;        //
        commandBlock[2] = 0;        //
        commandBlock[3] = 0;        //
        commandBlock[4] = 18;       // 蓖Ē
        commandBlock[5] = 0;        // Control

        errorCode = USBHostMSDRead( *address, 0, commandBlock, 6, inquiryData, 18 );
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PutChar( ' ' );
        #endif

        if (!errorCode)
        {
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

        #ifdef DEBUG_MODE
            UART2PrintString( "SCSI: READ CAPACITY 10..." );
        #endif

        // R}hubNREAD CAPACITY(10) p[^͂܂B
        // Xg[W̗eʏv܂B
        commandBlock[0] = 0x25;     // R[h
        commandBlock[1] = 0;        //
        commandBlock[2] = 0;        //
        commandBlock[3] = 0;        //
        commandBlock[4] = 0;        //
        commandBlock[5] = 0;        //
        commandBlock[6] = 0;        //
        commandBlock[7] = 0;        //
        commandBlock[8] = 0;        //
        commandBlock[9] = 0x00;     // Control

        errorCode = USBHostMSDRead( *address, 0, commandBlock, 10, inquiryData, 8 );
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PutChar( ' ' );
        #endif

        if (!errorCode)
        {
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

        if (!errorCode)
        {
            #ifdef DEBUG_MODE
                UART2PutHex( errorCode ) ;
                UART2PrintString( "\r\n" );
            #endif

            // ZN^[TCY肵܂B
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: Sector size:" );
                UART2PutChar( inquiryData[7] + '0' );
                UART2PutChar( inquiryData[6] + '0' );
                UART2PutChar( inquiryData[5] + '0' );
                UART2PutChar( inquiryData[4] + '0' );
                UART2PrintString( "\r\n" );
            #endif
            mediaInformation.sectorSize                     = (inquiryData[7] << 12) + (inquiryData[6] << 8) + (inquiryData[5] << 4) + (inquiryData[4]);
            mediaInformation.validityFlags.bits.sectorSize  = 1;

            mediaInformation.errorCode = MEDIA_NO_ERROR;
            return &mediaInformation;
        }
        else
        {
            // NA悤Ƃ邽߂ɁARequest Senses܂B
            #ifdef DEBUG_MODE
                UART2PrintString( "SCSI: REQUEST SENSE..." );
            #endif

            // R}hubNREQUEST SENSEp[^͂܂B
            commandBlock[0] = 0x03;     // R[h
            commandBlock[1] = 0;        //
            commandBlock[2] = 0;        //
            commandBlock[3] = 0;        //
            commandBlock[4] = 18;       // 蓖Ē
            commandBlock[5] = 0;        // Control

            errorCode = USBHostMSDRead( *address, 0, commandBlock, 6, inquiryData, 18 );
            #ifdef DEBUG_MODE
                UART2PutHex( errorCode ) ;
                UART2PutChar( ' ' );
            #endif

            if (!errorCode)
            {
                while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
                {
                    USBTasks();
                }
            }
        }
    }

    mediaInformation.errorCode = MEDIA_CANNOT_INITIALIZE;
    return &mediaInformation;

}

/****************************************************************************
  Function:
    bool USBHostMSDSCSIMediaDeinitialize( void * mediaConfig )

  Tv:
    ̋@\̓fBA񏉊܂B

  :
    ̋@\̓fBA񏉊܂B

  Precondition:
    None

  :
    mediaConfig - fBA\ւ̃|C^

  ߂l:
    true  - 
    false - ̑

  Remarks:
    None
  ***************************************************************************/
bool USBHostMSDSCSIMediaDeinitialize(void *mediaConfig)
{
    return true;
}

/****************************************************************************
  Function:
    bool USBHostMSDSCSIMediaReset( void  )

  Tv:
    ̋@\̓fBAZbg܂B

  :
    삪G[ԂꍇɌĂяo܂B 
    ́AAvP[VĂяoł܂B

  Precondition:
    None

  Parameters:
    None - None

  ߂l:
    USB_SUCCESS                 - Zbg
    USB_MSD_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_ILLEGAL_REQUEST         - Zbgׂ̈ɂ̓foCXsUSBԂɂȂĂ܂

  Remarks:
    None
  ***************************************************************************/

uint8_t USBHostMSDSCSIMediaReset( uint8_t * address )
{
    uint32_t   byteCount;
    uint8_t    errorCode;

    // oNp̃}XXg[WZbgJn܂B
    errorCode = USBHostMSDResetDevice( *address );
    if (errorCode)
    {
        return errorCode;
    }

    do
    {
        USBTasks();
        // }XXg[W̏Ԃ𓾂܂B
        errorCode = USBHostMSDDeviceStatus( *address );
    } while (errorCode == USB_MSD_RESETTING_DEVICE);

    // ]̊҂܂B
    if (USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
    {
        return errorCode;
    }

    return USB_MSD_RESET_ERROR;
}


/****************************************************************************
  Function:
    uint8_t USBHostMSDSCSISectorRead( uint32_t sectorAddress, uint8_t *dataBuffer)

  Tv:
    ̋@\1ZN^ǂݎ܂B

  :
    ̋@\SCSIR}hREAD(10)gp1ZN^ǂݎ܂B
    ZN^[̃TCY USBHostMSDSCSIMediaInitialize()֐Ō肳Ă܂B
    f[^̓AvP[Vobt@Ɋi[܂B

  Precondition:
    None

  :
    uint32_t   sectorAddress   - ǂݎZN^̃AhX
    uint8_t    *dataBuffer     - ǂݎf[^i[obt@̃|C^

  ߂l:
    true    - ǂݎ͐Ɏs܂
    false   - ǂݎ͐܂ł

  l:
    READ(10) - foCXf[^ǂݎ܂B
    READ10R}hubN͎̂ƂłB

    <code>
        Byte/Bit    7       6       5       4       3       2       1       0
           0                    Operation Code (0x28)
           1        [    RDPROTECT      ]  DPO     FUA      -     FUA_NV    -
           2        [ (MSB)
           3                        Logical Block Address
           4
           5                                                          (LSB) ]
           6        [         -         ][          Group Number            ]
           7        [ (MSB)         Transfer Length
           8                                                          (LSB) ]
           9        [                    Control                            ]
    </code>
  ***************************************************************************/

uint8_t USBHostMSDSCSISectorRead(uint8_t * address, uint32_t sectorAddress, uint8_t *dataBuffer )
{
    uint32_t   byteCount;
    uint8_t    commandBlock[10];
    uint8_t    errorCode;
    uint8_t    attempts;

    if (*address == 0)
    {
        return false;       // USB_MSD_DEVICE_NOT_FOUND;
    }

    attempts = INITIALIZATION_ATTEMPTS; // ǉAO5̐ݒłB
    while(attempts--)
    {
        // R}hubNREAD(10)p[^͂܂B
        commandBlock[0] = 0x28;     // R[h
        commandBlock[1] = RDPROTECT_NORMAL | FUA_ALLOW_CACHE;
        commandBlock[2] = (uint8_t) (sectorAddress >> 24);     // Big endian!
        commandBlock[3] = (uint8_t) (sectorAddress >> 16);
        commandBlock[4] = (uint8_t) (sectorAddress >> 8);
        commandBlock[5] = (uint8_t) (sectorAddress);
        commandBlock[6] = 0x00;     // O[vԍ
        commandBlock[7] = 0x00;     // ubN - Big endian!
        commandBlock[8] = 0x01;
        commandBlock[9] = 0x00;     // Control

        // LUN = 0gpĂ܂B
        // t@CVXeLUNT|[gĂꍇA͕ς܂B
        errorCode = USBHostMSDRead( *address, 0, commandBlock, 10, dataBuffer, mediaInformation.sectorSize );

        if (!errorCode)
        {   // ]̊҂܂B
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

        switch(errorCode)
        {
            case USB_MSD_COMMAND_FAILED:
                // ǂݏoR}h̃G[𓾂
                USBHostMSDSCSIRequestSense(address);
                break;

            case USB_SUCCESS:
                return true;

            default:
                return false;
        }
    }

    return false;
}

static bool USBHostMSDSCSIRequestSense(uint8_t * address)
{
    uint32_t   byteCount;
    uint8_t    commandBlock[10];
    uint8_t    errorCode;
    uint8_t buffer[18];

    if (*address == 0)
    {
        return false;       // USB_MSD_DEVICE_NOT_FOUND;
    }

    // R}hubNREAD10p[^͂܂B
    commandBlock[0] = 0x03;     // REQUEST SENSE
    commandBlock[1] = RDPROTECT_NORMAL | FUA_ALLOW_CACHE;
    commandBlock[2] = 0;     
    commandBlock[3] = 0;
    commandBlock[4] = 0x12;
    commandBlock[5] = 0;
    commandBlock[6] = 0x00;     // O[vԍ
    commandBlock[7] = 0x00;     // ubN - Big endian!
    commandBlock[8] = 0x00;
    commandBlock[9] = 0x00;     // Control

    // LUN = 0gpĂ܂B
    // t@CVXeLUNT|[gĂꍇA͕ς܂B
    errorCode = USBHostMSDRead( *address, 0, commandBlock, 10, buffer, 18 );

    if (!errorCode)
    {   // ]̊҂܂B
        while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
        {
            USBTasks();
        }
    }

    if (!errorCode)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/****************************************************************************
  Function:
    uint8_t USBHostMSDSCSISectorWrite( uint32_t sectorAddress, uint8_t *dataBuffer, uint8_t allowWriteToZero )

  Tv:
    ̊֐1ZN^݂܂B

  :
    ̋@\SCSIR}hWRITE10gp1ZN^݂܂B
    ZN^[̃TCY USBHostMSDSCSIMediaInitialize()֐Ō肳Ă܂B
    f[^̓AvP[Vobt@ǂݍ܂܂B

  Precondition:
    None

  :
    uint32_t   sectorAddress    - ރZN^̃AhX
    uint8_t    *dataBuffer      - AvP[Vf[^܂ރobt@̃|C^
    uint8_t    allowWriteToZero - ZN^0ւ݂̏Ăꍇ

  ߂l:
    true    - ݂͐Ɏs܂
    false   - ݂͐܂ł

  l:
    ɏ]āÅ֐͏݂܂ŃubN܂B
    WRITE(10) - foCXɃf[^݂܂B
    WRITE10R}hubN͎̂ƂłB

    <code>
        Byte/Bit    7       6       5       4       3       2       1       0
           0                    Operation Code (0x2A)
           1        [    WRPROTECT      ]  DPO     FUA      -     FUA_NV    -
           2        [ (MSB)
           3                        Logical Block Address
           4
           5                                                          (LSB) ]
           6        [         -         ][          Group Number            ]
           7        [ (MSB)         Transfer Length
           8                                                          (LSB) ]
           9        [                    Control                            ]
    </code>
  ***************************************************************************/

uint8_t USBHostMSDSCSISectorWrite(uint8_t * address, uint32_t sectorAddress, uint8_t *dataBuffer, uint8_t allowWriteToZero )
{
    uint32_t   byteCount;
    uint8_t    commandBlock[10];
    uint8_t    errorCode;
    uint8_t    attempts;

    if (*address == 0)
    {
        return false;   //USB_MSD_DEVICE_NOT_FOUND;
    }

    if ((sectorAddress == 0) && (allowWriteToZero == false))
    {
        return false;
    }

    attempts = INITIALIZATION_ATTEMPTS; // ǉAO5̐ݒłB
    while(attempts--)
    {
        // WRITE(10)p[^ŃR}hubN𖄂߂܂B
        commandBlock[0] = 0x2A;     // R[h
        commandBlock[1] = WRPROTECT_NORMAL | FUA_ALLOW_CACHE;
        commandBlock[2] = (uint8_t) (sectorAddress >> 24);     // Big endian!
        commandBlock[3] = (uint8_t) (sectorAddress >> 16);
        commandBlock[4] = (uint8_t) (sectorAddress >> 8);
        commandBlock[5] = (uint8_t) (sectorAddress);
        commandBlock[6] = 0x00;     // O[vԍ
        commandBlock[7] = 0x00;     // ubN - Big endian!
        commandBlock[8] = 0x01;
        commandBlock[9] = 0x00;     // Control

        // LUN = 0gpĂ܂B
        // t@CVXeLUNT|[gĂꍇA͕ς܂B
        errorCode = USBHostMSDWrite( *address, 0, commandBlock, 10, dataBuffer, mediaInformation.sectorSize );

        if (!errorCode)
        {   // ]̊҂܂B
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }

        switch(errorCode)
        {
            case USB_MSD_COMMAND_FAILED:
                // ݃R}h̃G[𓾂
                USBHostMSDSCSIRequestSense(address);
                break;

            case USB_SUCCESS:
                return true;

            default:
                return false;
        }
    }

    return false;
}


/****************************************************************************
  Function:
    uint8_t USBHostMSDSCSIWriteProtectState( void )

  :
    ̊֐̓foCX̃CgveNgXe[^XԂ܂B

  Precondition:
    None

  Parameters:
    None - None

  ߂l:
    0 - ݋֎~


  Remarks:
    None
  ***************************************************************************/

uint8_t    USBHostMSDSCSIWriteProtectState( uint8_t * address)
{
    return 0;
}


// *****************************************************************************
// *****************************************************************************
// Section: Internal Functions
// *****************************************************************************
// *****************************************************************************


/*******************************************************************************
  Function:
    bool _USBHostMSDSCSI_TestUnitReady( void )

  Precondition:
    None

  Overview:
    ̊֐TEST UNIT READY SCSIR}h𑗂܂B
    foCX]ɑΉ\ǂ₢킹܂B

  Parameters:
    None - None

  Return Values:
    true    - R}h̓G[ȂŊ܂
    false   - R}hs̃G[

  Remarks:
    TEST UNIT READYR}ȟ`͎̂ƂłB

    <code>
        Byte/Bit    7       6       5       4       3       2       1       0
           0                    Operation Code (0x00)
           1        [
           2                               Reserved
           3
           4                                                                ]
           5        [                    Control                            ]
    </code>
  ***************************************************************************/

#ifdef PERFORM_TEST_UNIT_READY
bool _USBHostMSDSCSI_TestUnitReady( uint8_t * address)
{
    uint32_t       byteCount;
    uint8_t        commandBlock[10];
    uint8_t        errorCode;
    uint8_t        inquiryData[36];
    uint8_t        unitReadyCount;

    // TEST UNIT READY𔭍s܂
    #ifdef DEBUG_MODE
        UART2PrintString( "SCSI: TEST UNIT READY..." );
    #endif

    unitReadyCount = 0;
    while (unitReadyCount < 5)
    {
        unitReadyCount ++;

        // R}hubNTEST UNIT READYp[^͂܂B
        commandBlock[0] = 0x00;     // R[h
        commandBlock[1] = 0;        // Reserved
        commandBlock[2] = 0;        // Reserved
        commandBlock[3] = 0;        // Reserved
        commandBlock[4] = 0;        // Reserved
        commandBlock[5] = 0x00;     // Control

        errorCode = USBHostMSDRead( *address, 0, commandBlock, 6, inquiryData, 0 );
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PutChar( ' ' );
        #endif

        if (!errorCode)
        {   // ]̊҂܂B
            while (!USBHostMSDTransferIsComplete( *address, &errorCode, &byteCount ))
            {
                USBTasks();
            }
        }
        #ifdef DEBUG_MODE
            UART2PutHex( errorCode ) ;
            UART2PrintString( "\r\n" );
        #endif
        if (!errorCode)
        {
            return true;
        }
    }

    return false;
}
#endif

